#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import unittest
from unittest.mock import patch, mock_open
import textwrap

import numpy as np

import craterstats as cst

class TestCratercount(unittest.TestCase):

    def test_ReadStatFile(self):
        f = r'd:\tmp\test.stat'
        c = textwrap.dedent("""
            # D        F(D)    N_dif    Error       C(D)    N_cum    Error 
            .0600       10  2.93E+00  9.28E-01      477  1.40E+00  6.41E-02
            .0700       12  3.52E+00  1.02E+00      467  1.37E+00  6.34E-02
            """)
        with patch('builtins.open', mock_open(read_data=c)):
            cc = cst.Cratercount(f)
        # check correctly read
        self.assertTrue(np.array_equal(cc.binned['d_min'],[0.06, 0.07]))
        self.assertTrue(np.array_equal(cc.binned['ncum'], [477, 467]))
        # check initialisation
        self.assertEqual(cc.binning,'unknown')
        self.assertTrue(cc.prebinned)
        # check derived values - tests MakeBinGeometricMean()
        self.assertAlmostEqual(cc.area, 477/1.40e+00) # area found from C(D)/N_cum
        self.assertTrue(np.allclose(cc.binned['d_max'], [0.07, .07*(.07/.06)])) # extrapolated last bin width
        self.assertAlmostEqual(cc.binned['d_mean'][0], np.sqrt(.06*.07))
        self.assertAlmostEqual(cc.binned['bin_width'][0], .07-.06)


    def test_ReadDiamFile1(self):
        c = textwrap.dedent("""
            # comment
            area = 100.
            crater = {diameter
            1.1
            2.3
            3.0
            1.4
            }
            """)
        with patch('builtins.open', mock_open(read_data=c)):
            cc = cst.Cratercount('test.diam')
        # check correctly read
        self.assertTrue(np.array_equal(cc.diam, [3., 2.3, 1.4, 1.1]))
        self.assertEqual(cc.area, 100.)
        # check initialisation
        self.assertEqual(cc.binning, None)
        self.assertFalse(cc.prebinned)
        self.assertFalse(cc.buffered)
        self.assertEqual(cc.perimeter, None)
        self.assertTrue(np.array_equal(cc.fraction, [1., 1., 1., 1.]))

    def test_ReadDiamFile2(self):
        c = textwrap.dedent("""
            area = 100.
            crater = {diameter, fraction, lon, lat
            1.1 0.3 127.3 -52.8
            }
            """)
        with patch('builtins.open', mock_open(read_data=c)):
            cc = cst.Cratercount('test.diam')
        # check correctly read (lat/lon not used here)
        self.assertEqual(cc.diam[0], 1.1)
        self.assertEqual(cc.fraction[0], .3)

    def test_ReadDiamFile3(self): # buffered count (indicated by presence of 'reference_area')
        c = textwrap.dedent("""
            crater = {diameter, reference_area
            1.1 24.3
            }
            """)
        with patch('builtins.open', mock_open(read_data=c)):
            cc = cst.Cratercount('test.diam')
        # check correctly read
        self.assertEqual(cc.diam[0], 1.1)
        # check initialisation
        self.assertEqual(cc.area, 1.) # crater densities relative to 1 km2 for buffered
        self.assertTrue(cc.buffered)
        # check derived values
        self.assertEqual(cc.fraction[0], 1/24.3)

    def test_ReadSCCfile(self):
        c = textwrap.dedent("""
            Total_area = 224669.2 <km^2>
            Perimeter = 1605.7
            # crater_diameters:
            crater = {diam,lon,lat
            33.58 -126.0 16.81
            40.40 -124.4 16.95
            }
            """)
        with patch('builtins.open', mock_open(read_data=c)):
            cc = cst.Cratercount('test.scc')
        # check correctly read
        self.assertTrue(np.array_equal(cc.diam, [33.58, 40.40]))
        self.assertEqual(cc.area, 224669.2)
        self.assertEqual(cc.perimeter, 1605.7)
        # check initialisation
        self.assertEqual(cc.binning, None)
        self.assertFalse(cc.prebinned)
        self.assertFalse(cc.buffered)
        self.assertTrue(np.array_equal(cc.fraction, [1., 1.]))

    def make_flat_distribution(self):
        self.N_CRATERS=100
        cc = cst.Cratercount('')
        cc.diam=10**np.linspace(.001,.999,self.N_CRATERS)
        cc.fraction=[1.]*(self.N_CRATERS-2)+[.3]*2
        cc.area = 10.
        return cc

    def test_apply_binning(self):
        cc=self.make_flat_distribution()
        cc.apply_binning('pseudo-log')
        self.assertEqual(cc.binned['d_min'][0], 1.)
        self.assertEqual(cc.binned['d_max'][-1], 10.)
        self.assertEqual(len(cc.binned['d_min']), 18)
        self.assertLess(max(cc.binned['n'])/min(cc.binned['n']), 3.4)
        self.assertEqual(cc.binned['ncum_event'][0], self.N_CRATERS)
        self.assertEqual(cc.binned['ncum'][0], self.N_CRATERS-1.4)
        self.assertAlmostEqual(cc.binned['d_mean'][0], np.sqrt(1.1))
        self.assertAlmostEqual(cc.binned['bin_width'][0], .1)
        cc.apply_binning('root-2')
        self.assertAlmostEqual(cc.binned['d_min'][1], np.sqrt(2))
        self.assertAlmostEqual(cc.binned['bin_width'][0], np.sqrt(2)-1)
        self.assertEqual(cc.binned['ncum_event'][0], self.N_CRATERS)
        self.assertLess(max(cc.binned['n'][0:-1])-min(cc.binned['n'][0:-1]), 2) # last bin may be cut
        b2=sum(cc.binned['n'][0:2])
        cc.apply_binning('x2')
        self.assertEqual(cc.binned['n'][0], b2) # encompasses 2 bins of previous
        self.assertLess(max(cc.binned['n'][0:-1]) - min(cc.binned['n'][0:-1]), 2)  # last bin may be cut
        cc.apply_binning('4th root-2')
        self.assertEqual(sum(cc.binned['n'][0:4]), b2)
        cc.apply_binning('10/decade')
        self.assertEqual(len(cc.binned['d_min']), 10)
        self.assertLess(max(cc.binned['n'][0:-1]) - min(cc.binned['n'][0:-1]), 2)  # last bin may be cut
        self.assertEqual(cc.binned['ncum_event'][0], self.N_CRATERS)
        cc.apply_binning('none')
        self.assertEqual(cc.binned['ncum_event'][0], self.N_CRATERS)
        self.assertEqual(cc.binned['ncum'][0], self.N_CRATERS-1.4)

    def test_getplotdata(self):
        cc = self.make_flat_distribution()
        p = cc.getplotdata('cumulative','none')
        self.assertEqual(p['bin_range'], (cc.diam[0],cc.diam[-1]))
        self.assertEqual(p['d'][0], cc.diam[0])
        self.assertEqual(p['y'][0], (self.N_CRATERS-1.4)/cc.area )
        p = cc.getplotdata('cumulative', 'pseudo-log',range=[1.02,4.98])
        self.assertEqual(p['bin_range'], (1.,5.))

    def test_generate_bins(self):
        cc = self.make_flat_distribution()
        self.assertEqual(cc.generate_bins('none',[.33,.44]),[.33,.44])
        np.testing.assert_allclose(cc.generate_bins('pseudo-log', [.33, .34]), [.3, .35])
        np.testing.assert_allclose(cc.generate_bins('pseudo-log', [.33, .35]), [.3, .35,.4])
        np.testing.assert_allclose(cc.generate_bins('pseudo-log', [.33, .35], expand=False), [.3, .35])
        np.testing.assert_allclose(cc.generate_bins('root-2', [1.1, 1.2]), [1., np.sqrt(2)])
        np.testing.assert_allclose(cc.generate_bins('root-2', [1.2, 1.3], offset=.5), [2**.25,2**.75])
        bins=cc.generate_bins('10/decade', cc.diam)
        self.assertEqual(len(bins),11)
        np.testing.assert_allclose(bins[[0,-1]],[1.,10.])
        self.assertEqual(len(cc.generate_bins('20/decade', cc.diam)),21)



if __name__ == '__main__':
    unittest.main()



import cli
import shlex
import os

def main():

    out = ' -o d:/mydocs/tmp/out.png'
    dir = ' -o d:/mydocs/tmp/'

    workdir = r"d:/mydocs/tmp"
    #workdir = r"D:\mydocs\publications\2024-10 sequence probability\figs\ce6 site seq"
    os.chdir(workdir)

    cmd = '--help'
    cmd = '-p source=sample/sample.scc'+out
    cmd = r'-cs MarsIvan2001 -p source="D:\mydocs\craterstatsII\sample\Pickering.scc" -p type=pois,range=[8,20]' + out
    cmd = r'-pr cumul -cs neukumivanov -p source=sample\sample.binned'
    cmd = r'-legend n#r -cs marsni2001 -p source=D:\mydocs\tmp\test_opencratertools\test1.scc -p type=bpoiss,range=[6,80],offset_age=[14,-14]'
    cmd = r'-legend n#rap -cs marsni2001 -o D:\mydocs\data\congzhe\greg -p src="D:\mydocs\data\congzhe\greg\scarp1.scc" -p type=bp,range=[3,20]'
    cmd = r'-cs marsni2001 -p src=D:\mydocs\data\congzhe\greg\uplift.scc -p type=bpoi,range=[.6,1.2] -p range=[4,20]'
    cmd = r'-cs Moonv2 -pr cum -pt_size 3.6 -xrange -3 4 -yrange -8 7 -p src="D:\mydocs\work\_collaborators\astrid oetting\2024-05-11 my calculations\non-mare NSC.scc" -p type=c-fit,range=[100,200],isochron=1,age_left=1'
    cmd = r'-cs MoonN2001 -pr cum -pt_size 3.6 -xrange -3 4 -yrange -8 7 -p src="D:\mydocs\work\_collaborators\astrid oetting\2024-05-11 my calculations\non-mare NSC.scc" -p type=c-fit,range=[100,200],isochron=1,age_left=1'
    cmd = r'-cs Moonv2 -pr cum  -xrange 0 5 -yrange -8 -3 -p src="D:\mydocs\work\_collaborators\astrid oetting\2024-05-07 lunar basin pf\New_PF_Basin_Full_Moon_ex_Mare.scc" -p type=c-fit,range=[100,200],isochron=1 -p range=[600,1000] -p range=[1500,2500]'
    cmd = r'-cs N83 -xrange -3 0 -sf 2 -isochrons .05 -p src="D:\mydocs\work\_gyig\li yang\2024-06-04 ce6 landing\ce6_50m_crater.scc" -p type=bp,range=[.01,.060]' + dir
    cmd = r'-i d:/mydocs/tmp/ce6_50m_crater.cs'
    cmd = r'-cs neukumivanov -ep mars -ef standard -p source=%sample%/Pickering.scc -p type=poisson,range=[2,5],offset_age=[2,-2] -p range=[.2,.7]'+dir
    cmd = r'-pr sq -cs yue22 -ep moon -xrange 0 4.5 -p source=%sample%/Pickering.scc -p type=poisson,range=[.2,7] -p range=[2,5] -p range=[.1,.2],colour=red -p range=[.5,.7] -p type=d-fit,range=[2,5],colour=black -p name=test2,range=[.2,.7] -p range=[.1,.2] -p range=[.5,.7]' + dir
    cmd = r'-i D:\mydocs\tmp\Pickering.cs'
    cmd = r'-i barbara.cs'

    cmd = r'-i "D:\mydocs\publications\2024-10 sequence probability\figs\ce6 site seq\ce-6 mare W5.cs"'
    cmd = r'-cs neukumivanov -title Example plot -p source=%sample%/Pickering.scc -p type=poisson,range=[b8,b-5]'

    cmd = r'-cs neukumivanov -title Example plot -p source=%sample%/Pickering.scc -p type=poisson,range=[b8,b-5]'
    cmd = r'--functions_user "D:\mydocs\code\pycharm\craterstats\src\craterstats\config\functions_user.txt"'
    cmd = r'-i "D:\mydocs\tmp\demo\00-demo.cs"'

    #cmd = r'-pr cumul -cs neukumivanov -pd 8x8'+dir

    #cmd = '--help'

    #cmd = r'-demo'



    #-print_dimensions 7.5x7.5 -pt_size 8

    print(f'\nDebugging command: craterstats '+cmd)
    a = shlex.split(cmd)
    cli.main(a)

if __name__ == '__main__':
    main()


#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

__version__ = "3.1.1"

from .Chronologyfn import Chronologyfn
from .Productionfn import Productionfn
from .Cratercount import Cratercount
from .Craterplotset import Craterplotset
from .Craterplot import Craterplot
from .Craterpdf import Craterpdf
from .Epochs import Epochs

from .bin_bias_correction import bin_bias_correction
from .str_age import str_age

from .constants import *






#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

from .normal import normal
from .poisson import poisson
from .range import range,mag
from .poly import poly
from .scl import scl



#  Copyright (c) 2021, Greg Michael
#  Licensed under BSD 3-Clause License. See LICENSE.txt for details.

import sys

import numpy as np
import math
from scipy.integrate import simps

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker

import craterstats as cst
import craterstats.gm as gm


class Craterpdf:
    """
    Create and plot age uncertainty distribution function

    """

    def __init__(self, pf,cf,cc,d_range,k=None,lam=None,bcc=False,n_samples=5000):
        """
        :param pf: Productionfn object
        :param cf: Chronologyfn object
        :param cc: Cratercount object
        :param d_range: diameter range (km)
        :param k: number of craters in range (normally not required - calculated from cc)
        :param lam: lambda for 1 Ga (if wish to force value, e.g. for probability calculation)
        :param bcc: buffered crater count?
        :param n_samples: number of samples for likelihood curve
        """

        x=np.linspace(-10, 5, n_samples) #additive change to a0 (equidistant in fitting space)
        a0=cf.a0(1.)
        self.ts = cf.t(a0=a0+x)
        self.dt = self.ts - np.roll(self.ts, 1)
        self.dt[0] = self.dt[1]

        if k is not None:
            self.k=k
        elif cc.diam:
            d = [e for e in cc.diam if d_range[0] <= e < d_range[1]]
            self.k = len(d)
        else:   # we have a binned-only count
            b = cc['binned']
            q = np.where(b['d_min'] >= d_range[0] and b['d_max'] <= d_range[1])
            self.k = np.sum(b['.n_event'][q])

        if not lam:
            if bcc: #buffered count
                if cc.perimeter is None:
                    sys.exit('Error: buffered-poisson calculation requires polygon perimeter in source file')
                ns=500 #samples of PF for integration
                d=np.linspace(d_range[0], d_range[1], ns)
                F=pf.F(d,a0)
                y=(cc.area+d*cc.perimeter/2.+np.pi*d**2/8.)*F
                I1=simps(y,d)
                lam=I1
            else: #standard count
                Ncum=pf.evaluate("cumulative",d_range,a0)
                Ninc=(Ncum[0]-Ncum[1])*cc.area # no expected on area with phi(1 Ga)
                lam=Ninc

        pdf0 = gm.poisson(self.k, lam * 10 ** x)
        pdf0 = pdf0.astype(float)
        self.lam = lam
        self.pdf = pdf0/np.sum(pdf0*self.dt)
        self.cdf = np.cumsum(self.pdf*self.dt)


    def t(self,cum_fraction):
        """
        Return time for interpolated percentiles

        :param cum_fraction: percentile as fraction
        :return: times
        """
        return np.interp(cum_fraction,self.cdf,self.ts)

    def cumulative_fraction(self,t):
        """
        Return interpolated percentile for time

        :param t: time, Ga
        :return: cum_fraction
        """
        return np.interp(t,self.ts,self.cdf)

    def relative_probability(self,t):
        """
        Return relative_probability for time

        :param t: time, Ga
        :return: relative_probability
        """
        return np.interp(t,self.ts,self.pdf)

    def gaussian_percentiles(self,n=1):
        """
        Return ordered Gaussian n-sigma percentiles

        :param n: max n-sigma
        :return: ordered percentiles (as fractions)
        """
        g= [.5]
        for i in range(1,n+1):
            f = (1-math.erf(i/np.sqrt(2.)))/2
            g = [f]+g+[1-f]
        return g

    def median1sigma(self):
        """
        Return times for median and 1-sigma percentiles

        :return: times
        """
        g=self.gaussian_percentiles()
        return self.t([g[i] for i in [1,0,2]])


    def plot(self,ax=None,pt_size=9,color='0',t_range=[],logscale=False, max_ticks=3):
        """
        Set up uncertainty distribution plot

        :param ax: matplotlib axes object
        :param pt_size: character point size
        :param color: colour
        :return: none
        """
        if not ax:
            ax = plt.subplot(1,1,1)

        linewidth=.5 * pt_size / 9
        t=self.t([.003]+self.gaussian_percentiles()+[.997])
        p=np.searchsorted(self.ts,t)-1

        if not logscale:
            if t_range:
                max_t = t_range[1]
            else:
                max_t = np.max(t)
            xt = gm.ticks(np.array([0.,max_t]), max_ticks)
            #2024-07-25 don't understand what this was for:
            # for i,e in enumerate(xt):
            #     if e<=max_t: max_i=i
            max_i=-1

            max_text = cst.str_age(xt[max_i], simple=True)
            a = max_text.split(' ')
            xt_units = float(a[0]) / xt[max_i] * xt
            xt_label = ['{:g}'.format(e) for e in xt_units]
            xt_label[max_i] = "      " + xt_label[max_i] + " " + a[1]  # add unit to last label, e.g. "Ga"

        else: #logscale
            if not t_range:
                t_range=gm.range(t)
                xt0=list(range(np.floor(np.log10(t_range[0])).astype(int),np.ceil(np.log10(t_range[1])).astype(int)+1))
            else:
                xt0= ([np.log10(t_range[0])]
                    +list(range(np.floor(np.log10(t_range[0])).astype(int)+1,np.ceil(np.log10(t_range[1])).astype(int)))
                    +[np.log10(t_range[1])])

            xt = [10 ** e for e in xt0]
            xt_label = [cst.str_age(e, simple=True) for e in xt]
            ax.set_xscale('log')


        ax.plot(self.ts, self.pdf, lw=linewidth * 1.5, color=color)
        ax.fill_between(self.ts[p[1]:p[3]], self.pdf[p[1]:p[3]], 0, color=color, alpha=0.35, edgecolor=color,
                        lw=linewidth)  # '.7'
        ax.fill_between(self.ts[p[2]:p[2] + 1], self.pdf[p[2]:p[2] + 1], edgecolor=color, lw=linewidth, alpha=.5)

        ax.patch.set_facecolor('none') # make region transparent over background graphics
        ax.get_yaxis().set_visible(False)
        for e in ['right','left','top']: ax.spines[e].set_visible(False)
        ax.spines['bottom'].set_linewidth(linewidth)
        ax.spines['bottom'].set_color(color)

        ax.set_xticks(xt)
        ax.set_xlim(xt[0], t_range[1] if t_range else xt[-1],auto=False)

        ax.tick_params(axis='x', which='both', width=linewidth, length=pt_size * .2, pad=pt_size * .1, color=color)
        ax.tick_params(axis='x', which='minor', length=pt_size * .1)
        if not logscale: ax.xaxis.set_minor_locator(ticker.AutoMinorLocator())
        ax.set_xticklabels(xt_label,fontsize=pt_size*.7, color=color)#,horizontalalignment='left')

    def offset(self,left):
        """
        Calculate offset for correct spacing of age text from mini-plot peak, in +ve x-direction

        :param left: left-aligned plot?
        :return: offset as fraction of mini-plot width
        """
        t=self.t(.997)
        max_ticks=3
        xt= gm.ticks(np.append(0, t), max_ticks)
        q=np.where(self.pdf > np.max(self.pdf)/2)
        edges=self.ts[[q[0][0],q[0][-1]]]
        norm=edges/np.max(xt)
        offset=(1-norm[1])-.1 if left else -norm[0]+.1
        return offset

    def violin_plot(self,ax,y,w0,pt_size=9,color='0',invert=False):
        """
        Set up uncertainty distribution plot

        :param ax: matplotlib axes object
        :param pt_size: character point size
        :param color: colour
        :return: none
        """
        linewidth=.4 * pt_size / 9
        t=self.t([.0001,.16,.5,.84,.9999])
        p=np.searchsorted(self.ts,t)-1

        w = .9*w0/max(self.pdf)

        ax.fill_between(self.ts[p[0]:p[4]], y + w * self.pdf[p[0]:p[4]], y - w * self.pdf[p[0]:p[4]],
                        color='black' if invert else 'white',
                        edgecolor=color, lw=linewidth * 1.5, zorder=4)

        ax.fill_between(self.ts[p[1]:p[3]], y + w * self.pdf[p[1]:p[3]], y - w * self.pdf[p[1]:p[3]],
                        color=gm.mix_colours(color,'black' if invert else 'white',0.35),
                        edgecolor=color,lw=linewidth, zorder=4)

        ax.fill_between(self.ts[p[2]:p[2] + 1], y + w * self.pdf[p[2]:p[2] + 1], y - w * self.pdf[p[2]:p[2] + 1],
                        edgecolor=color, lw=linewidth, alpha=.5, zorder=4)


    def calculate_sequence_probability(self,pdf2):
        """
        Calculate sequence probability with another Craterpdf object
        
        :param pdf2: another Craterpdf object
        :return: probability that self older than pdf
        """
        # Discretisation of cdf biases result, since calculated for one edge, especially for narrow pdfs.
        # Finding midpoint to fix:
        cdf2=(pdf2.cdf+np.roll(pdf2.cdf, 1))/2.
        P = np.sum(self.dt * self.pdf * cdf2)
        return P

    def calculate_instantaneous_probability_ratio(self, t):
        """
        Calculate probability ratio between pdf median and another time, t

        :param t: other time, t
        :return: probability ratio pr(t)/pr(median) [usually <1. if median close to max]
        """
        v = np.interp([self.t(0.5),t], self.ts, self.pdf)
        r = v[1]/v[0]
        return r

    # def calculate_mean_relative_likelihood(self, t):
    #     """
    #     experimental
    #     want mean relative likelihood over pdf if given median t (area stays same, n now unknown)
    #
    #
    #     :return: probability that self older than pdf
    #     """
    #     #have to create new pdf, since lambda changed...
    #
    #     mrl = np.sum(pdf2**2/median_pdf2*pdf2.dt)
    #
    #     return mrl